# SPDX-License-Identifier: BSD-2-Clause
""" Builds documents for the Sphinx documentation framework. """

# Copyright (C) 2020, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from contextlib import contextmanager
import logging
import os
import re
import shutil
from typing import Any, Callable, Iterator, Optional
import yaml

from specitems import (BibTeXCitationProvider, Copyrights,
                       DocumentGlossaryConfig, GlossaryConfig, Item,
                       ItemGetValueContext, ItemMapper, SphinxContent,
                       generate_glossary, get_value_subprocess, is_enabled,
                       list_terms, to_iterable)
from specware import BSD_2_CLAUSE_LICENSE, run_command

from .directorystate import DirectoryState
from .pkgitems import (BuildItem, BuildItemFactory, BuildItemMapper,
                       PackageBuildDirector, PackageComponent)
from .testoutputparser import augment_report

_BREAK = "\\break"

_PUSH_ENABLED_BY = re.compile(r"^\${\.:/push-enabled-by:(.+)}$")

_POP_ENABLED_BY = re.compile(r"^\${\.:/pop-enabled-by")

_RST_HEADERS = re.compile(
    r"^\.\. SPDX-License-Identifier: (.+)\n\n((\.\. Copyright \(C\).*\n)+)\n",
    flags=re.MULTILINE)

_INVISIBLE_SPACES_AT = re.compile(r"([\/_-]+)")


def spacify(text: str) -> str:
    """ Add invisible spaces to enable line breaks. """
    return _INVISIBLE_SPACES_AT.sub("\u200b\\1", text)


def _normal_title(item: Item) -> str:
    return item["document-title"].replace(_BREAK, " ")


def _get_normal_title(ctx: ItemGetValueContext) -> str:
    return ctx.substitute_and_transform(_normal_title(ctx.item))


def _get_sphinx_title(ctx: ItemGetValueContext) -> str:
    content = SphinxContent()
    content.add_header(_get_normal_title(ctx), level=1)
    return content.join().rstrip()


def _get_release(ctx: ItemGetValueContext) -> str:
    return str(len(ctx.item["document-releases"]))


def _no_action(_source_dir: str, _build_dir: str,
               _component: dict[str, Any]) -> None:
    pass


def _sep(seps: tuple[str, ...], maxi: tuple[int, ...]) -> str:
    return "+" + "+".join(f"{sep * (val + 2)}"
                          for sep, val in zip(seps, maxi)) + "+"


def _row(row: tuple[str, ...], maxi: tuple[int, ...]) -> str:
    return "|" + "|".join(f" {cell:{width}} "
                          for cell, width in zip(row, maxi)) + "|"


def _get_contributors(ctx: ItemGetValueContext) -> str:
    rows = [("Action", "Name", "Organization", "Signature")]
    maxi = tuple(map(len, rows[0]))
    for action in ctx.mapper.substitute_data(
            ctx.item["document-contributors"]):
        for contributor in action["contributors"]:
            first = contributor["first-name"]
            last = contributor["last-name"]
            row = (action["action"], f"{first} {last}",
                   contributor["organization"], "")
            rows.append(row)
            row_lengths = tuple(map(len, row))
            maxi = tuple(map(max, zip(maxi, row_lengths)))
    sep_0 = _sep(("-", "-", "-", "-"), maxi)
    sep_1 = _sep(("=", "=", "=", "="), maxi)
    sep_2 = _sep((" ", "-", "-", "-"), maxi)
    lines = [sep_0, _row(rows[0], maxi)]
    last_action = rows[0][0]
    for row in rows[1:]:
        if last_action == "Action":
            lines.append(sep_1)
            last_action = row[0]
        elif last_action == row[0]:
            lines.append(sep_2)
            row = ("", row[1], row[2], row[3])
        else:
            lines.append(sep_0)
            last_action = row[0]
        lines.append(_row(row, maxi))
    lines.append(sep_0)
    content = SphinxContent()
    with content.directive(
            "table", options=[":class: longtable", ":widths: 16 26 30 28"]):
        content.add(lines)
    return content.join()


def _latex_escape(value: str) -> str:
    return value.replace("_", "\\_").replace("&", "\\&")


_COPYRIGHT = re.compile(r"^\s*Copyright\s+\(C\)\s+", re.IGNORECASE)
_YEARS = re.compile(r"^[0-9, ]*")


def _document_copyright(item: Item, mapper: ItemMapper) -> str:
    copyrights = item["document-copyrights"]
    main = _COPYRIGHT.sub("", copyrights[0])
    main = mapper.substitute(main)
    if len(copyrights) == 1:
        return main
    return f"{main} and contributors"


def _document_author(item: Item, mapper: ItemMapper) -> str:
    return _YEARS.sub("", _document_copyright(item, mapper))


def _document_year(item: Item, mapper: ItemMapper) -> str:
    match = _YEARS.search(_document_copyright(item, mapper))
    assert match
    return match.group(0).split(",")[-1].strip()


def _get_document_copyright(ctx: ItemGetValueContext) -> str:
    return _document_copyright(ctx.item, ctx.mapper)


def _get_document_author(ctx: ItemGetValueContext) -> str:
    return _document_author(ctx.item, ctx.mapper)


def _get_document_year(ctx: ItemGetValueContext) -> str:
    return _document_year(ctx.item, ctx.mapper)


def _get_latex_title(ctx: ItemGetValueContext) -> str:
    return _latex_escape(
        ctx.substitute_and_transform(ctx.item["document-title"].replace(
            _BREAK, " ").replace("\\strut", " ").replace("\\hfill", " ")))


def _get_value_latex(ctx: ItemGetValueContext) -> str:
    return _latex_escape(ctx.substitute_and_transform(ctx.value[ctx.key]))


def _get_latex_copyright(ctx: ItemGetValueContext) -> str:
    return _latex_escape(_get_document_copyright(ctx))


def _get_title_page_title(ctx: ItemGetValueContext) -> str:
    title: list[str] = []
    for part in ctx.substitute_and_transform(
            ctx.item["document-title"]).split(_BREAK):
        if len(part) < 23:
            title.append(part)
            continue
        parts = part.split("/")
        new_part = parts[0]
        for sub_part in parts[1:]:
            sub_part = f"/{sub_part}"
            if len(new_part) + len(sub_part) < 23:
                new_part += sub_part
            else:
                title.append(new_part)
                new_part = sub_part
        title.append(new_part)
    return _latex_escape(" \\break \\break ".join(title))


def _gather_subcomponents(component: Item, depth: int, at_depth: int,
                          component_list: list[str]) -> None:
    for subcomponent in component.children("input"):
        if not subcomponent.type.startswith("pkg/component"):
            continue
        if depth == at_depth:
            name = os.path.dirname(component.uid)
            subname = os.path.dirname(subcomponent.uid)
            component_list.append(subname.removeprefix(name).lstrip("/"))
        else:
            _gather_subcomponents(subcomponent, depth + 1, at_depth,
                                  component_list)


def _augment_report(
        lines: list[str]) -> tuple[list[str], list[tuple[int, int]]]:
    report: dict = {}
    augment_report(report, lines)
    return lines, report["data-ranges"]


class _CitationProvider(BibTeXCitationProvider):

    def __init__(self, builder: "SphinxBuilder") -> None:
        super().__init__(builder.mapper)
        self.add_get_fields("pkg/directory-state/sphinx",
                            self._get_document_fields)
        self.mapper.add_get_value("pkg/directory-state/sphinx:/bibtex-entries",
                                  self.get_bibtex_entries)
        for type_name in ("pkg/component", "pkg/directory-state/sphinx",
                          "statement-of-compliance"):
            self.mapper.add_get_value(f"{type_name}:/cite-group",
                                      self.get_cite_group)

    def _get_document_fields(
            self, item: Item) -> tuple[str, dict[str, str | list[str]]]:
        organizations: set[str] = set()
        authors: set[str] = set()
        for action in item["document-contributors"]:
            for contributor in action["contributors"]:
                first = contributor["first-name"]
                last = contributor["last-name"]
                authors.add(f"{last}, {first}")
                organizations.add(contributor["organization"])
        fields: dict[str, str | list[str]] = {
            "author": sorted(authors),
            "organization": ", ".join(sorted(organizations)),
            "title": _normal_title(item),
            "url": f"{item['directory']}/{item['output-pdf']}",
            "year": _document_year(item, self.mapper)
        }
        return "manual", fields


class SphinxBuilder(DirectoryState):
    """ Base class for Sphinx document builds. """

    @classmethod
    def prepare_factory(cls, factory: BuildItemFactory,
                        type_name: str) -> None:
        DirectoryState.prepare_factory(factory, type_name)
        factory.add_get_value(f"{type_name}:/document-release", _get_release)
        factory.add_get_value(f"{type_name}:/document-year",
                              _get_document_year)

    def __init__(self,
                 director: PackageBuildDirector,
                 item: Item,
                 mapper: Optional[BuildItemMapper] = None) -> None:
        super().__init__(director, item, mapper)
        _CitationProvider(self)
        self._index: list[str] = []
        self._section_level_stack: list[int] = [2]
        self._section_stack: list[BuildItem] = [self]
        self.content_license: set[str] = {self["document-license"]}
        for source_license in self["document-license-map"].keys():
            for the_license in source_license.split(" OR "):
                self.content_license.add(the_license)
        self.file_path = ""
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/subcomponent-list",
                                  self._get_subcomponent_list)
        self.mapper.add_get_value(f"{my_type}:/document-author",
                                  _get_document_author)
        self.mapper.add_get_value(f"{my_type}:/document-latex-footer-right",
                                  _get_value_latex)
        self.mapper.add_get_value(f"{my_type}:/document-latex-header-left",
                                  _get_value_latex)
        self.mapper.add_get_value(f"{my_type}:/document-copyright",
                                  _get_document_copyright)
        self.mapper.add_get_value(f"{my_type}:/document-copyrights",
                                  self._get_document_copyrights)
        self.mapper.add_get_value(
            f"{my_type}:/document-bsd-2-clause-copyrights",
            self._get_document_bsd_2_clause_copyrights)
        self.mapper.add_get_value(f"{my_type}:/document-normal-title",
                                  _get_normal_title)
        self.mapper.add_get_value(f"{my_type}:/document-latex-copyright",
                                  _get_latex_copyright)
        self.mapper.add_get_value(f"{my_type}:/document-latex-title",
                                  _get_latex_title)
        self.mapper.add_get_value(f"{my_type}:/document-sphinx-title",
                                  _get_sphinx_title)
        self.mapper.add_get_value(f"{my_type}:/document-title-page-title",
                                  _get_title_page_title)
        self.mapper.add_get_value(f"{my_type}:/document-index",
                                  self._get_index)
        self.mapper.add_get_value(f"{my_type}:/document-releases",
                                  self._get_release_topics)
        self.mapper.add_get_value(f"{my_type}:/document-release-sections",
                                  self._get_release_sections)
        self.mapper.add_get_value(f"{my_type}:/document-contributors",
                                  _get_contributors)
        for name in [my_type, "pkg/sphinx-section"]:
            self.mapper.add_get_value(f"{name}:/build-description",
                                      self._get_build_description)
            self.mapper.add_get_value(f"{name}:/document-elements",
                                      self._get_document_elements)
            self.mapper.add_get_value(f"{name}:/push-component",
                                      self._push_component)
            self.mapper.add_get_value(f"{name}:/pop-component",
                                      self._pop_component)
            self.mapper.add_get_value(f"{name}:/sections", self._get_sections)
            self.mapper.add_get_value(f"{name}:/subprocess",
                                      self._get_subprocess)
        self.mapper.add_get_value("pkg/sphinx-section:/ref",
                                  self._get_section_ref)
        self._actions = {
            "add-to-index": _no_action,
            "copy": self._copy,
            "copy-and-substitute": self._copy_and_substitute,
            "glossary": self._glossary
        }

    def _run_actions(self, source: DirectoryState, build_dir: str) -> None:
        source_dir = source.directory
        document_components = self["document-components"]
        files = set(source.files(base=""))
        for component in document_components:
            files.difference_update(to_iterable(component.get("file",
                                                              tuple())))
        for file in sorted(files):
            self._do_copy(source_dir, build_dir, file)
        enabled_set = self.enabled_set
        for component in document_components:
            if is_enabled(enabled_set, component.get("enabled-by", True)):
                self._actions[component["action"]](source_dir, build_dir,
                                                   component)
                self._add_to_index(component)

    def run(self) -> None:
        self.mapper.copyrights_by_license.clear()

        source = self.input("source")
        assert isinstance(source, DirectoryState)

        build = self.output("build")
        assert isinstance(build, DirectoryState)
        build_dir = build.directory

        try:
            destination = self.output("destination")
            assert isinstance(destination, DirectoryState)
        except KeyError:
            destination = self
        destination.discard()
        destination.clear()

        os.makedirs(os.path.join(build_dir, "source"), exist_ok=True)
        status = run_command(
            ["python3", "-msphinx", "-M", "clean", "source", "build"],
            build_dir)
        assert status == 0
        self._run_actions(source, build_dir)
        output = self["output-pdf"]
        if output:
            stdout: list[str] = []
            status = run_command(
                ["python3", "-msphinx", "-M", "latexpdf", "source", "build"],
                build_dir,
                stdout=stdout)
            unstable = ("Latexmk: Maximum runs of pdflatex reached"
                        " without getting stable files")
            assert status == 0 or unstable in stdout
            src_path = os.path.join(build_dir, "build", "latex",
                                    "document.pdf")
            destination.copy_file(src_path, output)
        output = self["output-html"]
        if output:
            status = run_command(
                ["python3", "-msphinx", "-M", "html", "source", "build"],
                build_dir)
            assert status == 0
            src_path = os.path.join(build_dir, "build", "html")
            destination.copy_tree(src_path, output)

        my_license = self["document-license"]
        destination["copyrights-by-license"] = dict(
            (key, value.get_statements())
            for key, value in self._get_copyrights().items()
            if key != my_license)

        self.description.add(f"""Produce documents in
{self.description.path(destination.directory)} using sources from
{source.reference()}.""")

    def add_component_action(
            self, name: str, action: Callable[[str, str, dict[str, Any]],
                                              None]) -> None:
        """ Adds a component action. """
        self._actions[name] = action

    def wrap(self, content: SphinxContent, item: Item, text: str) -> None:
        """ Substitute the text and wrap the text to the content. """
        content.wrap(self.mapper.substitute(text, item))

    def _get_subcomponent_list(self, ctx: ItemGetValueContext) -> str:
        assert ctx.args
        at_depth = int(ctx.args)
        component_list: list[str] = []
        _gather_subcomponents(self.director[self.director.package_uid].item, 1,
                              at_depth, component_list)
        return list_terms(sorted(component_list))

    def _get_releases(self, ctx: ItemGetValueContext, scope: Callable) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(self.section_level)
            releases = ctx.item["document-releases"]
            count = len(releases)
            for idx, release in enumerate(reversed(releases)):
                date = release["date"]
                status = release["status"]
                header = ctx.substitute(
                    f"Release: {count - idx}, Date: {date}, Status: {status}")
                with scope(content, header):
                    lines = self._push_pop_enabled_by(
                        release["changes"].splitlines())
                    content.add(ctx.substitute("\n".join(lines)))
            return content.join()

    def _get_release_topics(self, ctx: ItemGetValueContext) -> str:

        @contextmanager
        def _scope(content: SphinxContent, header: str) -> Iterator[None]:
            with content.directive("topic", header):
                yield

        return self._get_releases(ctx, _scope)

    def _get_release_sections(self, ctx: ItemGetValueContext) -> str:

        @contextmanager
        def _scope(content: SphinxContent, header: str) -> Iterator[None]:
            with content.section(header):
                yield

        return self._get_releases(ctx, _scope)

    def _add_to_index(self, component: dict[str, Any]) -> None:
        if component.get("add-to-index", False):
            for file in to_iterable(component["file"]):
                self._index.append(os.path.basename(file).replace(".rst", ""))

    def _get_index(self, ctx: ItemGetValueContext) -> str:
        content = SphinxContent()
        maxdepth = f":maxdepth: {ctx.item['document-toctree-maxdepth']}"
        with content.directive("toctree", options=[maxdepth, ":numbered:"]):
            content.add(self._index)
        return content.join()

    def _get_copyrights(self) -> dict[str, Copyrights]:
        my_license = self["document-license"]
        copyrights: dict[str, Copyrights] = {}
        copyrights.setdefault(my_license, Copyrights()).register(
            self["document-copyrights"])
        license_map = self["document-license-map"]
        for key, statements in self.mapper.copyrights_by_license.items():
            the_license = license_map.get(key, key)
            copyrights.setdefault(the_license, Copyrights()).register([
                license_map.get(statement, statement)
                for statement in statements
            ])
        return copyrights

    def _get_document_copyrights(self, ctx: ItemGetValueContext) -> str:
        my_license = self["document-license"]
        assert " OR " not in my_license
        copyrights = self._get_copyrights()
        prefix = ctx.args if ctx.args else ""
        return "\n".join(copyrights[my_license].get_statements(f"{prefix}| ©"))

    def _get_document_bsd_2_clause_copyrights(
            self, _ctx: ItemGetValueContext) -> str:
        copyrights = self._get_copyrights()
        the_license = "BSD-2-Clause"
        if the_license not in copyrights:
            return ""
        statements = "\n".join(copyrights[the_license].get_statements("| ©"))
        return f"""{statements}

{BSD_2_CLAUSE_LICENSE}"""

    @property
    def section_level(self) -> int:
        """ Is the current section level. """
        return self._section_level_stack[-1]

    @contextmanager
    def section_level_scope(
            self, ctx: ItemGetValueContext) -> Iterator[Optional[str]]:
        """ Opens a section level scope with optional additional arguments. """
        if ctx.args:
            colon = ctx.args.find(":")
            if colon >= 0:
                level_change = int(ctx.args[:colon])
                args: str | None = ctx.args[colon + 1:]
            else:
                level_change = int(ctx.args)
                args = None
        else:
            level_change = 1
            args = None
        self._section_level_stack.append(self.section_level + level_change)
        yield args
        self._section_level_stack.pop()

    def _add_section_content(self, content: SphinxContent,
                             section: BuildItem) -> None:
        lines = self._push_pop_enabled_by(
            section.item["content"].strip().splitlines())
        self._section_stack.append(section)
        content.add(self.substitute("\n".join(lines), section.item))
        self._section_stack.pop()

    def _push_component(self, ctx: ItemGetValueContext) -> str:
        assert ctx.args
        component = self.director[ctx.args]
        assert isinstance(component, PackageComponent)
        self.push_component(component)
        item_cache = self.item.cache
        item_cache.push_selection(component.selection)
        item_cache.push_view(component.view)
        return ""

    def _pop_component(self, _ctx: ItemGetValueContext) -> str:
        item_cache = self.item.cache
        item_cache.pop_view()
        item_cache.pop_selection()
        self.pop_component()
        return ""

    def _get_section_label(self, item: Item) -> str:
        label = item["label"]
        if label is None:
            return item.ident
        return self.substitute(label, item)

    def _get_sections(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx) as section_key:
            content = SphinxContent(self.section_level)
            assert section_key
            links_sections = [
                (link, section)
                for link, section in self._section_stack[-1].input_links(
                    "document-section") if link["section-key"] == section_key
            ]
            if len(links_sections
                   ) == 1 and links_sections[0][0]["header-is-optional"]:
                section = links_sections[0][1]
                self._section_level_stack[-1] = self._section_level_stack[-2]
                logging.debug("%s: add content %s at level %s using %s",
                              self.uid, section.uid, self.section_level,
                              section.component.uid)
                with self.component_scope(section.component):
                    content.add_label(self._get_section_label(section.item))
                    self._add_section_content(content, section)
            else:
                for link, section in links_sections:
                    logging.debug("%s: add section %s at level %s using %s",
                                  self.uid, section.uid, self.section_level,
                                  section.component.uid)
                    with self.component_scope(section.component):
                        with content.section(
                                self.substitute(link["header"], section.item),
                                label=self._get_section_label(section.item)):
                            self._add_section_content(content, section)
            return content.join()

    def _get_section_ref(self, ctx: ItemGetValueContext) -> str:
        section = self.director[ctx.item.uid]
        with self.component_scope(section.component):
            return f":ref:`{self._get_section_label(section.item)}`"

    def _get_document_elements(self, ctx: ItemGetValueContext) -> str:
        assert ctx.args
        indent, element_key = ctx.args.split(":")
        content = SphinxContent()
        content.push_indent(int(indent) * " ")
        elements = [
            element for element in self.inputs("document-element")
            if element["element-key"] == element_key
        ]
        if len(elements) == 1:
            element = elements[0]
            logging.debug("%s: add element %s using %s", self.uid, element.uid,
                          element.component.uid)
            with self.component_scope(element.component):
                content.add(self.substitute(element.item["content"]))
        else:
            for element in elements:
                logging.debug("%s: add element %s using %s", self.uid,
                              element.uid, element.component.uid)
                with self.component_scope(element.component):
                    content.add_list_item(
                        self.substitute(element.item["header"]))
                    content.add_blank_line()
                    with content.indent("  "):
                        content.add(self.substitute(element.item["content"]))
        return content.join()

    def _get_subprocess(self, ctx: ItemGetValueContext) -> str:
        return get_value_subprocess(self.substitute, _augment_report, ctx)

    def _get_build_description(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx) as args:
            assert args
            content = SphinxContent(self.section_level)
            self.director.add_build_description(content, args.split(":"))
            return content.join()

    def _register_text_copyrights(self, text: str) -> None:
        match = _RST_HEADERS.match(text)
        assert match
        the_license = match.group(1)
        statements = [
            statement[3:] for statement in match.group(2).split("\n")[:-1]
        ]
        logging.info("%s: register license %s with copyrights %s", self.uid,
                     the_license, statements)
        self.mapper.copyrights_by_license.setdefault(the_license,
                                                     set()).update(statements)

    def _do_copy(self, source_dir: str, build_dir: str, file: str) -> None:
        src_file = os.path.join(source_dir, file)
        dst_file = os.path.join(build_dir, file)
        os.makedirs(os.path.dirname(dst_file), exist_ok=True)
        if dst_file.endswith(".rst"):
            logging.info("%s: read: %s", self.uid, src_file)
            with open(src_file, "r", encoding="utf-8") as src:
                text = src.read()
                self._register_text_copyrights(text)
                logging.info("%s: write: %s", self.uid, dst_file)
                with open(dst_file, "w+", encoding="utf-8") as dst:
                    dst.write(text)
        else:
            logging.info("%s: copy '%s' to '%s'", self.uid, src_file, dst_file)
            shutil.copy2(src_file, dst_file)

    def _copy(self, source_dir: str, build_dir: str,
              component: dict[str, Any]) -> None:
        for file in to_iterable(component["file"]):
            self._do_copy(source_dir, build_dir, file)

    def _push_pop_enabled_by(self, lines: list[str]) -> list[str]:
        logging.info("%s: process push/pop enabled by using enabled set: %s",
                     self.uid, sorted(self.enabled_set))
        filtered_lines: list[str] = []
        enabled: list[bool] = [True]
        for line in lines:
            push_match = _PUSH_ENABLED_BY.search(line)
            if push_match is not None:
                data = push_match.group(1)
                enabled_by = yaml.load(data, yaml.SafeLoader)
                enabled.append(enabled[-1]
                               and is_enabled(self.enabled_set, enabled_by))
                continue
            if _POP_ENABLED_BY.search(line) is not None:
                enabled.pop()
            elif enabled[-1]:
                filtered_lines.append(line)
        return filtered_lines

    def _copy_and_substitute(self, source_dir: str, build_dir: str,
                             component: dict[str, Any]) -> None:
        for file in to_iterable(component["file"]):
            src_path = os.path.join(source_dir, file)
            dst_path = os.path.join(build_dir, file)
            self.file_path = dst_path
            logging.info("%s: read: %s", self.uid, src_path)
            with open(src_path, "r", encoding="utf-8") as src:
                component_depth = self.component_depth
                text = "".join(self._push_pop_enabled_by(src.readlines()))
                if dst_path.endswith(".rst"):
                    self._register_text_copyrights(text)
                logging.info("%s: substitute", self.uid)
                text = self.mapper.substitute(text)
                logging.info("%s: write: %s", self.uid, dst_path)
                os.makedirs(os.path.dirname(dst_path), exist_ok=True)
                with open(dst_path, "w+", encoding="utf-8") as dst:
                    dst.write(text)
                assert component_depth == self.component_depth

    def _glossary(self, _source_dir: str, build_dir: str,
                  component: dict[str, Any]) -> None:
        document_config = DocumentGlossaryConfig(
            target=str(os.path.join(build_dir, component["file"])),
            header="Terms, definitions and abbreviated terms",
            rest_source_paths=[
                str(os.path.join(build_dir, "source")),
                str(os.path.join(build_dir, "include"))
            ])
        config = GlossaryConfig(project_groups=component["glossary-groups"],
                                documents=[document_config])
        generate_glossary(config, self.item.cache, self.mapper, SphinxContent)



# SPDX-License-Identifier: BSD-2-Clause
""" Maintains an item cache directory state. """

# Copyright (C) 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from typing import Optional

from specitems import Item, ItemCache, ItemCacheConfig, ItemTypeProvider
from specware import augment_with_test_case_links, augment_with_test_links

from .directorystate import DirectoryState
from .pkgitems import PackageBuildDirector


class ItemCacheDirectoryState(DirectoryState):
    """ Maintains an item cache directory state. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self._cache: ItemCache | None = None

    def load(self) -> str:
        self.cache()
        return super().load()

    def spec_paths(self, base: Optional[str] = None) -> list[str]:
        """ Return the specification item paths. """
        if base is None:
            source = self.input("source")
            assert isinstance(source, DirectoryState)
            base = source.directory
        return [os.path.join(base, path) for path in self["spec-paths"]]

    def cache(self) -> ItemCache:
        """ Return the item cache. """
        item_cache = self._cache
        if item_cache is None:
            cache_config = ItemCacheConfig(paths=self.spec_paths(),
                                           cache_directory=self.directory)
            type_provider = self.item.cache.type_provider
            item_cache = ItemCache(cache_config,
                                   type_provider=ItemTypeProvider(
                                       type_provider.data_by_uid,
                                       type_provider.root_type_uid,
                                       permissive_type_errors=True))
            augment_with_test_case_links(item_cache)
            augment_with_test_links(item_cache)
            self._cache = item_cache
        return item_cache



# SPDX-License-Identifier: BSD-2-Clause
""" Generates a command to run executables. """

# Copyright (C) 2024, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import stat
import subprocess

from .directorystate import DirectoryState
from .pkgitems import Redirection
from .testrunner import TestLog, TestRunner

_COMMAND_HEAD = """#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
\"\"\" Runs the executable on the target. \"\"\"

# Copyright (C) 2024 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import base64
import json
import hashlib
import os
import subprocess
import sys
"""

_COMMAND_TAIL = """def _hash_file(path: str) -> str:
    file_hash = hashlib.sha512()
    if os.path.islink(path):
        file_hash.update(os.readlink(path).encode("utf-8"))
    else:
        buf = bytearray(65536)
        memview = memoryview(buf)
        with open(path, "rb", buffering=0) as src:
            for size in iter(lambda: src.readinto(memview), 0):  # type: ignore
                file_hash.update(memview[:size])
    return base64.urlsafe_b64encode(file_hash.digest()).decode("ascii")


def main(script: str, argv: list[str]) -> int:
    \"\"\" Runs the executables on the target. \"\"\"
    package_directory = os.path.normpath(
        os.path.join(os.path.dirname(__file__), ".."))
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--runner-type",
        default="simulator",
        type=str,
        help=f"the type of the executable runner (default: simulator)")
    parser.add_argument(
        "--package-directory",
        default=package_directory,
        type=str,
        help=f"the package directory (default: {package_directory})")
    parser.add_argument("--do-not-use-test-logs",
                        action="store_true",
                        help=f"do not use the test logs to look up a report")
    parser.add_argument("target",
                        metavar="TARGET",
                        nargs=1,
                        help="the target on which the executables should run; "
                        "the target is the name of an installed BSP "
                        f"(one of {list(_TARGETS_BY_BSP.keys())}) "
                        f"or a runner (one of {list(_COMMANDS.keys())})")
    parser.add_argument("executables",
                        metavar="EXECUTABLES",
                        nargs="+",
                        help="the executable files to run")
    args = parser.parse_args(argv)
    target = args.target[0]
    target = _TARGETS_BY_BSP.get(target, target)
    try:
        command = _COMMANDS[target]
    except KeyError:
        command = _COMMANDS.get(f"{target}/{args.runner_type}", None)
    reports_by_hash: dict = {}
    if not args.do_not_use_test_logs:
        for test_log_file in _TEST_LOGS_BY_TARGET.get(target, []):
            path = test_log_file.replace("${package-directory}",
                                         args.package_directory)
            try:
                with open(path, "r", encoding="utf-8") as file:
                    for report in json.load(file)["reports"]:
                        digest = report["executable-sha512"]
                        if digest not in reports_by_hash:
                            reports_by_hash[digest] = report["output"]
            except FileNotFoundError:
                pass
    for executable in args.executables:
        digest = _hash_file(executable)
        report = reports_by_hash.get(digest, None)
        if report is None:
            if command is None:
                print(
                    f"{script}: in your environment, "
                    "there is no service available to run the executable "
                    f"on the target '{target}'; "
                    f"available runner targets are: {list(_COMMANDS.keys())} ",
                    file=sys.stderr)
                return 1
            command = [
                arg.replace("${executable}",
                            executable).replace("${package-directory}",
                                                args.package_directory)
                for arg in command
            ]
            subprocess.run(command, check=True, stdin=subprocess.DEVNULL)
        else:
            print("\\n".join(report))
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[0], sys.argv[1:]))
"""


class RunExecutableCommand(DirectoryState):
    """ Generates a command to run executables. """

    def run(self) -> None:
        self.set_files([self["file-name"]])
        file_path = self.file
        logging.info("%s: create run executable command: %s", self.uid,
                     file_path)
        target_by_bsp: dict[str, str] = {}
        test_logs_by_target: dict[str, list[str]] = {}
        deployment_directory = self.substitute(
            "${.:/component/deployment-directory}")
        for test_log in self.inputs("test-log"):
            assert isinstance(test_log, TestLog)
            with self.component_scope(test_log.component):
                target = self._target()
                test_logs_by_target.setdefault(target, []).append(
                    test_log.file.replace(deployment_directory,
                                          "${package-directory}"))
                target_by_bsp[self.substitute(
                    "${.:/component/arch}-${.:/component/bsp}"
                    "${.:/component/config:dash}${.:/component/bsp-extra:dash}"
                )] = target
                target_by_bsp[self.substitute(
                    "${.:/component/arch}-${.:/component/bsp}"
                    "${.:/component/config:dash}"
                    "${.:/component/bsp-qual-only:dash}")] = target
        commands_by_runner: dict[str, list[str]] = {}
        for redirection in self.inputs("test-runner"):
            assert isinstance(redirection, Redirection)
            with self.component_scope(redirection.component):
                test_runner = redirection.target()
                assert isinstance(test_runner, TestRunner)
                with test_runner.component_scope(redirection.component):
                    command = test_runner.get_run_command("$${executable}")
                    if command is None:
                        continue
                    command = [
                        arg.replace(deployment_directory,
                                    "${package-directory}") for arg in command
                    ]
                    runner_type = redirection["params"]["test-runner-type"]
                    runner = f"{self._target()}/{runner_type}"
                    assert runner not in commands_by_runner
                    commands_by_runner[runner] = command
                    logging.info("%s: add command %s for runner %s", self.uid,
                                 command, runner)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(_COMMAND_HEAD)
            file.write(f"""_TARGETS_BY_BSP: dict[str, str] = {target_by_bsp}

_TEST_LOGS_BY_TARGET: dict[str, str] = {test_logs_by_target}

_COMMANDS: dict[str, list[str]] = {commands_by_runner}
""")
            file.write(_COMMAND_TAIL)
        os.chmod(file_path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        subprocess.run(["yapf", "-i", file_path], check=True)

        self.description.add(f"""Produce the test run script
file {self.description.path(file_path)}.""")

    def _target(self) -> str:
        return self.substitute("${.:/component/arch}/${.:/component/bsp}"
                               "${.:/component/config:slash}")



# SPDX-License-Identifier: BSD-2-Clause
""" Functions for Building RTEMS SPAMR document. """

# Copyright (C) 2020 EDISOFT
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=consider-using-f-string

import os
import ast
import glob
import logging
from typing import Any, List, Tuple

from specitems import Item, ItemCache, ItemGetValueContext, SphinxContent
from specware import run_command

from .directorystate import DirectoryState
from .docbuilder import DocumentBuilder
from .pkgitems import PackageBuildDirector
from .rtems import RTEMSItemCache


def _get_specs_by_type(item_cache: ItemCache, spec_type: str,
                       enabled_by: List) -> List[Item]:
    """
    Returns a list of all enabled specifications of a certain type
    which are present on the given item cache.
    """
    specs = []
    for spec in item_cache.values():
        if spec.type.startswith(spec_type) and spec.is_enabled(enabled_by):
            specs.append(spec)
    return specs


def _get_test_cases_by_type(item_cache: ItemCache, enabled_by: Any,
                            test_type: str) -> List[Item]:
    """ Returns only the test cases of a given type. """
    test_cases = []  # type: List[Item]
    for case in _get_specs_by_type(item_cache, "test-case", enabled_by):
        links = list(case.parents(test_type))
        if links:
            test_cases.append(case)
    return sorted(test_cases)


class SpamrManager(DocumentBuilder):
    """ Class which manages the SPAMR document build step """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self.mapper.add_get_value(f"{self.item.type}:/cppcheck-results",
                                  self._cppcheck_analysis)
        self.mapper.add_get_value(
            f"{self.item.type}:/" +
            "requirement-implementation-and-vv-coverage",
            self._requirement_implementation_and_vv_coverage)
        self.mapper.add_get_value(
            f"{self.item.type}:/requirements-completeness",
            self._requirements_completeness)
        self.mapper.add_get_value(
            f"{self.item.type}:/vv-coverage-unit-integration-level",
            self._vv_coverage_unit_integration)
        self.mapper.add_get_value(f"{self.item.type}:/requirement-testability",
                                  self._requirement_testability)
        self.mapper.add_get_value(f"{self.item.type}:/code-metrics",
                                  self._code_metrics)
        self.mapper.add_get_value(
            f"{self.item.type}:/user-documentation-completeness",
            self._user_documentation_completeness)
        self.mapper.add_get_value(f"{self.item.type}:/code-size",
                                  self._code_size)
        self.mapper.add_get_value(f"{self.item.type}:/sprs-and-ncrws",
                                  self._sprs_and_ncrws)
        self.mapper.add_get_value(f"{self.item.type}:/spr-ncr-status",
                                  self._spr_ncr_status)

    def _get_related_requirements(self) -> list[Item]:
        spec = self.input("spec")
        assert isinstance(spec, RTEMSItemCache)
        return spec.get_related_requirements()

    def _cppcheck_analysis(self, _ctx: ItemGetValueContext) -> Any:
        command: List[str] = []
        for argument in self.item["cppcheck-command"]:
            command.append(self.mapper.substitute(argument))
        stdout: List[str] = []
        run_command(command, stdout=stdout)
        content = SphinxContent()
        for line in stdout:
            if not (("Checking " in line) or (" files checked " in line)):
                with content.directive("code-block", value="bash"):
                    content.add(
                        line.replace(self.component["deployment-directory"],
                                     ""))
        return content.join()

    def _requirement_implementation_and_vv_coverage(
            self, _ctx: ItemGetValueContext) -> Any:
        reqs_with_val_metho = 0
        requirements = self._get_related_requirements()
        number_requirements = len(requirements)
        not_validated_requirements = []
        for req in requirements:
            if req.view["validated"]:
                reqs_with_val_metho += 1
            else:
                not_validated_requirements.append(req.uid)
        report = []
        report.append("Number of software requirements with " +
                      "an associated validation method: " +
                      str(reqs_with_val_metho) + "\n")
        report.append("Total Number of software requirements: " +
                      str(number_requirements) + "\n")
        req_imp_cov = reqs_with_val_metho / number_requirements
        report.append("Requirement implementation coverage: " +
                      "{:.2f}".format(req_imp_cov) + "\n")
        if req_imp_cov >= self.item["metrics-thresholds"][
                "requirement-implementation-coverage"]:
            report.append("This metric is OK")
        else:
            report.append("This metric is NOK\n")
            report.append("The following requirements are not validated:\n")
            for requirement_uid in sorted(not_validated_requirements):
                report.append(requirement_uid + "\n")

        content = SphinxContent()
        content.add(report)
        return content.join()

    def _requirement_testability(self, _ctx: ItemGetValueContext) -> Any:
        reqs_with_val_test = 0
        requirements = self._get_related_requirements()
        number_requirements = len(requirements)
        for item in requirements:
            if item.type in [
                    "requirement/functional/action",
                    "requirement/non-functional/performance-runtime",
                    "runtime-measurement-test"
            ]:
                reqs_with_val_test += 1
                continue
            try:
                item.child("validation")
                reqs_with_val_test += 1
            except IndexError:
                pass
        report = []
        report.append("Number of software requirements verified by test: " +
                      str(reqs_with_val_test) + "\n")
        report.append("Total Number of software requirements: " +
                      str(number_requirements) + "\n")
        req_testability = reqs_with_val_test / number_requirements
        report.append("Requirement testability: " +
                      "{:.2f}".format(req_testability) + "\n")
        if req_testability * 100 >= self.item["metrics-thresholds"][
                "requirement-testability"]:
            report.append("This metric is OK")
        else:
            report.append("This metric is NOK")

        content = SphinxContent()
        content.add(report)
        return content.join()

    def _get_tbd_requirements(self) -> List[Item]:
        tbd_reqs = []
        for req in self._get_related_requirements():
            tbd_check_command = ["egrep", "-wq", "TBD|TBC", req.file]
            if run_command(tbd_check_command, stdout=[]) == 0:
                tbd_reqs.append(req)
        return tbd_reqs

    def _requirements_completeness(self, _ctx: ItemGetValueContext) -> Any:
        content = SphinxContent()
        number_tbds = len(self._get_tbd_requirements())
        requirement_count = len(self._get_related_requirements())
        req_comp = number_tbds / requirement_count
        content.add("Number of software requirements with TBCs/TBDs: " +
                    str(number_tbds))
        content.add(
            f"Total Number of software requirements: {requirement_count}")
        content.add("Requirements completeness: " + "{:.2f}".format(req_comp) +
                    " %")
        if req_comp <= \
                self.item["metrics-thresholds"]["requirements-completeness"]:
            content.add("This metric is OK")
        else:
            content.add("This metric is NOK")
        return content.join()

    def _user_documentation_completeness(self,
                                         _ctx: ItemGetValueContext) -> Any:
        build = self.output("build")
        assert isinstance(build, DirectoryState)
        build_dir = build.directory

        content = SphinxContent()
        sections: List[str] = []
        command_section = ["grep", r"\*\*\*\*\*$", "filename"]
        tbd_docs: List[str] = []
        command_tbd = ["grep", "TB[CD]", "filename"]

        # Search through the ReST documentation files
        for directory_state in self.inputs("rst-files"):
            assert isinstance(directory_state, DirectoryState)
            files = sorted(
                glob.glob(os.path.join(directory_state.directory, "**/*.rst")))
            logging.debug("%s: inspect documentation files in '%s': %s",
                          self.uid, directory_state.directory, files)
            for rst_file in files:
                command_section[2] = rst_file
                command_tbd[2] = rst_file
                tbd_file_doc: List[str] = []
                run_command(command_section, stdout=sections)
                run_command(command_tbd, stdout=tbd_file_doc)
                tbd_docs.extend([
                    os.path.relpath(rst_file, build_dir) + ": " + line
                    for line in tbd_file_doc
                ])

        # Write number of sections in RTEMS documentation
        content.add("Total number of sections in the documentation:")
        content.add_blank_line()
        content.add_list_item("sphinx documentation in RTEMS repository: " +
                              str(len(sections)))

        # Get number of TBD/TBC in requirements and documentation
        content.add_blank_line()
        content.add("Total number of TBC/TBD in the documentation:")
        content.add_blank_line()
        tbd_reqs = self._get_tbd_requirements()
        content.add_list_item("yml documentation (requirements): " +
                              str(len(tbd_reqs)))
        content.add_list_item("sphinx documentation in RTEMS repository: " +
                              str(len(tbd_docs)))
        content.add_blank_line()

        # Calculate user documentation completeness
        completeness = (len(tbd_docs) + len(tbd_reqs)) / len(sections)
        content.add(("User Documentation Completeness: (%d + %d) / %d = " %
                     (len(tbd_reqs), len(tbd_docs), len(sections))) +
                    "{:.2f}".format(completeness))
        content.add_blank_line()

        # Get relevant TBC/TBD lines
        content.add("Relevant list of TBC/TBD:")
        content.add_blank_line()
        if len(tbd_docs) != 0:
            content.add_list(tbd_docs)
            content.add_list(["Requirement: " + req.uid for req in tbd_reqs])
        else:
            content.add("No TBD/TBC found.")

        # Verify if metric is OK
        if completeness <= (self.item["metrics-thresholds"]
                            ["user-documentation-completeness"]):
            content.add("This metric is OK")
        else:
            content.add("This metric is NOK")
        return content.join()

    # pylint: disable=too-many-locals
    # pylint: disable=too-many-nested-blocks
    def _code_metrics(self, _ctx: ItemGetValueContext) -> Any:
        all_metrics = ""
        directory = ""
        # Run the metrics
        command = []
        for argument in self.item["metrixplusplus-collect"]:
            command.append(self.mapper.substitute(argument))
        run_command(command)
        # Gather and process the metrics
        command = self["metrixplusplus-view"]
        directory = command[-1]
        incorrect_metrics_files = []
        for root, _, files in os.walk(directory):
            for file in files:
                if file.endswith(".c") or file.endswith(".h"):
                    command[-1] = os.path.join(root, file)
                    stdout: List[str] = []
                    if run_command(command, stdout=stdout) == 0:
                        metrics, metrics_result, metrics_file = \
                            self._process_metrics(stdout, directory)
                        if len(metrics) != 1:
                            if not metrics_result:
                                incorrect_metrics_files.append(metrics_file)
                            content = SphinxContent()
                            with content.directive("code-block", value="text"):
                                content.add(metrics)
                            all_metrics = all_metrics + "\n\n" + content.join()

        wrong_files_content = SphinxContent()
        wrong_files_content.add("The following files contain errors in the " +
                                "the metrics. For more details see the " +
                                "above report of each file.\n\n")
        wrong_files_content.add_list(incorrect_metrics_files)

        return all_metrics + "\n\n" + wrong_files_content.join()

    def _process_metrics(self, metrics_output: List[str],
                         base_dir: str) -> Tuple[List[str], bool, str]:
        results_output = []
        for line in metrics_output:
            # Only one entry of the stdout will have the metrics dictionary
            if line.startswith("{"):
                while True:
                    if line.endswith("}"):
                        break
                    line = line[:-1]
                metrics_dict = ast.literal_eval(line)
        data = metrics_dict["view"][0]["data"]
        file_name = data["info"]["path"].replace(base_dir, "")
        results_output.append("File: " + file_name)
        metrics_ok = True
        code_size_file = int(
            data["aggregated-data"]["std.code.lines"]["code"]["total"] +
            data["aggregated-data"]["std.code.lines"]["preprocessor"]["total"])
        for region in data["file-data"]["regions"]:
            if region["info"]["type"] == "global":
                results_output[0] = results_output[0] + " " + str(
                    code_size_file)
                if code_size_file > self.item["metrics-thresholds"][
                        "lines-of-code"]:
                    overall_result = "NOK"
                    metrics_ok = False
                else:
                    overall_result = "OK"
                results_output[0] = results_output[0] + " " + overall_result
            if region["info"]["type"] == "function":
                cyclomatic = region["data"]["std.code.complexity"][
                    "cyclomatic"]
                nesting_level = region["data"]["std.code.complexity"][
                    "maxindent"]
                code_size = region["data"]["std.code.lines"]["code"] + region[
                    "data"]["std.code.lines"]["preprocessor"]
                comment_size = region["data"]["std.code.lines"]["comments"]
                comment_frequency = int(
                    (100 * comment_size) / (comment_size + code_size))
                if (cyclomatic > self.item[
                        "metrics-thresholds"]["cyclomatic-number"]) or \
                        (nesting_level > self.item[
                            "metrics-thresholds"]["nesting-level"]):
                    overall_result = "NOK"
                    metrics_ok = False
                else:
                    overall_result = "OK"
                results = "    " + region["info"]["name"] + " " + \
                          str(cyclomatic) + " " + str(nesting_level) + " " + \
                          str(code_size) + " " + str(comment_frequency) + \
                          "%" + " " + overall_result
                results_output.append(results)
        return results_output, metrics_ok, file_name

    def _code_size(self, _ctx: ItemGetValueContext) -> Any:
        # Action status
        sphinx_content = SphinxContent()
        actions = ["Actions issued", "Actions closed", "Actions open"]
        actions_count = [0, 0, 0]
        for issue in sorted(self.item.cache.values()):
            if not (issue.type == "issue" and "ESA" in issue["origin"]
                    and issue["issue-type"] == "MAI"):
                continue
            actions_count[0] += 1
            if issue["status"] == "closed":
                actions_count[1] += 1
            else:
                actions_count[2] += 1
        sphinx_content.add("Since this metric refers to project actions, "
                           "it will only consider the actions in Gitlab, "
                           "specific for the RTEMS SMP project.")
        sphinx_content.add("Below is presented the action status "
                           "(issues with label MAI in Gitlab):")
        for index, title in enumerate(actions):
            sphinx_content.add_list_item(title + ": " +
                                         str(actions_count[index]))
        sphinx_content.add_blank_line()
        # Code size
        command = self["metrixplusplus-view"]
        stdout: List[str] = []
        run_command(command, stdout=stdout)
        for line in stdout:
            # Only one entry of the stdout will have the metrics dictionary
            if line.startswith("{"):
                while True:
                    if line.endswith("}"):
                        break
                    line = line[:-1]
                metrics_dict = ast.literal_eval(line)
        info_dict = metrics_dict["view"][0]["data"]["aggregated-data"][
            "std.code.lines"]
        code_size = int(info_dict["code"]["total"] +
                        info_dict["preprocessor"]["total"])
        sphinx_content.add("The RTEMS code size is " + str(code_size))
        return sphinx_content.join()

    def _sprs_and_ncrws(self, _ctx: ItemGetValueContext) -> str:
        sphinx_content = SphinxContent()
        existing_rids = False
        sphinx_content.add("The following RIDs are currently open:")
        for issue in sorted(self.item.cache.values()):
            if not (issue.type == "issue" and issue["status"] != "closed"
                    and issue["issue-type"] == "RID"):
                continue
            existing_rids = True
            name = issue["origin"] + " #" + issue["issue-number"]
            url = f"`{name.strip()} <{issue['url']}>`_"
            with sphinx_content.indent():
                sphinx_content.add_list_item(url)
        output = {
            False: "There are no RIDs currently open.",
            True: sphinx_content.join()
        }
        return output[existing_rids]

    def _spr_ncr_status(self, _ctx: ItemGetValueContext) -> str:
        sphinx_content = SphinxContent()
        actions = ["Actions issued", "Actions closed", "Actions open"]
        actions_gitlab = [0, 0, 0]
        actions_rtems = [0, 0, 0]
        for issue in sorted(self.item.cache.values()):
            if not issue.type == "issue":
                continue
            action_origin = actions_gitlab if "ESA" in issue["origin"] \
                else actions_rtems
            action_origin[0] += 1
            if issue["status"] == "closed":
                action_origin[1] += 1
            else:
                action_origin[2] += 1
        sphinx_content.add(
            "Below is presented the action status for ESA Gitlab issues:")
        for index, title in enumerate(actions):
            sphinx_content.add_list_item(title + ": " +
                                         str(actions_gitlab[index]))
        sphinx_content.add_blank_line()
        sphinx_content.add_blank_line()
        sphinx_content.add("Below is presented the action status"
                           " for RTEMS Ticket System issues:")
        for index, title in enumerate(actions):
            sphinx_content.add_list_item(title + ": " +
                                         str(actions_rtems[index]))
        return sphinx_content.join()

    def _vv_coverage_unit_integration(self, _ctx: ItemGetValueContext) -> str:
        sphinx_content = SphinxContent()
        test_cases = _get_test_cases_by_type(self.item.cache, self.enabled_set,
                                             "unit-test")
        spec = self.input("spec")
        assert isinstance(spec, RTEMSItemCache)

        covered = set()
        modules = set(name for name in spec.name_to_item.keys()
                      if name.startswith("function/"))
        for item in test_cases:
            for link in item.links_to_parents("unit-test"):
                covered.add(link["name"])
        coverage = round((len(covered) / len(modules)) * 100)
        sphinx_content.add(f"Modules covered by V&V: {len(covered)}")
        sphinx_content.add(f"Total number of modules: {len(modules)}")
        sphinx_content.add(
            f"V&V coverage: {len(covered)} / {len(modules)} = {coverage} %")
        metric_ok = {True: "This metric is OK.", False: "This metric is NOK."}
        sphinx_content.add(
            metric_ok[self.item["metrics-thresholds"]["vv-coverage"] *
                      100 <= coverage])
        return sphinx_content.join()



# SPDX-License-Identifier: BSD-2-Clause
"""
Provides a command line interface to remove existing coverage files and run
gcovr using test output.
"""

# Copyright (C) 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import base64
import glob
import os
import re
import shutil
import subprocess
import sys

_GCOV_INFO = re.compile(r"\*\*\* BEGIN OF GCOV INFO BASE64 \*\*\*(.*)"
                        r"\*\*\* END OF GCOV INFO BASE64 \*\*\*")


def cligcovr(argv: list[str] = sys.argv) -> None:
    """ Removes existing coverage files and runs gcovr using test output. """
    parser = argparse.ArgumentParser()
    parser.add_argument("-o",
                        "--object-directory",
                        type=str,
                        default="build",
                        help="the path to the object directory")
    parser.add_argument(
        "-d",
        "--destination-directory",
        help="the path to the coverage report destination directory",
        default="coverage")
    parser.add_argument("-w",
                        "--working-directory",
                        help="the path to the working directory",
                        default=".")
    parser.add_argument("-r",
                        "--gcovr",
                        help="the path to the gcovr executable",
                        default="gcovr")
    parser.add_argument("-g",
                        "--gcov",
                        help="the path to the gcov executable",
                        default="gcov")
    parser.add_argument("-t",
                        "--gcov-tool",
                        help="the path to the gcov-tool executable",
                        default="gcov-tool")
    parser.add_argument("-v",
                        "--verbose",
                        action="store_true",
                        help="be verbose")
    parser.add_argument("outputs",
                        metavar="OUTPUTS",
                        nargs="+",
                        help="the test output files")
    args = parser.parse_args(argv[1:])
    for gcda in glob.glob(os.path.join(args.object_directory, "**/*.gcda"),
                          recursive=True):
        print("remove:", gcda)
        os.unlink(gcda)
    for output in args.outputs:
        with open(output, "r", encoding="latin-1") as src:
            lines = src.readlines()
            mobj = _GCOV_INFO.search("".join(line.strip() for line in lines))
            if mobj:
                gcov_info = base64.b64decode(mobj.group(1))
                cmd = [args.gcov_tool, "merge-stream"]
                print(f"run in {args.working_directory}: {' '.join(cmd)}")
                subprocess.run(cmd,
                               check=True,
                               input=gcov_info,
                               cwd=args.working_directory)
    shutil.rmtree(args.destination_directory, ignore_errors=True)
    os.makedirs(args.destination_directory)
    cmd = [
        "gcovr", "--include-internal-functions", "--json-summary",
        f"{args.destination_directory}/summary.json", "--json-summary-pretty",
        f"--gcov-executable={args.gcov}", "--html-details",
        f"{args.destination_directory}/index.html", "--html-medium-threshold",
        "80.0", "--html-high-threshold", "100.0", "--object-directory",
        args.object_directory, args.object_directory
    ]
    if args.verbose:
        cmd.insert(1, "--verbose")
    print(f"run in {args.working_directory}: {' '.join(cmd)}")
    subprocess.run(cmd, check=True, cwd=args.working_directory)
